/**
 * @file aufgabe02.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Schreiben Sie eine Methode, die alle Lösung des 8-Dame Problems ausgibt. Das 8-Dame Problem besteht darin, auf einem Schachbrett (8x8 Feld) 8 Damen derart zu platzieren, dass keine Dame eine andere Dame schlagen kann, d.h. in keiner Zeile und in keiner Spalte und in keiner Diagonalen dürfen 2 oder mehr Damen gleichzeitig stehen. (Tipp: arbeiten Sie rekursiv)
 * @version 0.1
 * @date 2022-05-14
 * 
 * @copyright Copyright (c) 2022
 */

#include <iostream>
using namespace std;

unsigned count = 0; 

/**
 * @brief Überprüft, ob eine Queen an der Stelle qx, qy auf dem Schachbrett board legal gestellt werden kann. Dabei wird nur diagonal oberhalb und vertikal ober- und unterhalb geprüft (heißt, die Positionen der Queens sollten von oben nach unten gesetzt/geprüft werden)
 * 
 * @param board Repräsentation des Schachbretts als eindimensionales int-Array
 * @param qx Zu prüfende x-Koordinate der Queen
 * @param qy Zu prüfende y-Koordinate der Queen
 * @return true Die geprüfte Position ist legal
 * @return false Die geprüfte Position ist nicht legal
 */
bool check(int board[8], int qx, int qy) {
    // ty = zu testendes y
    // board[ty] = zu testendes x
    // qy = y-Position der Queen
    // qx = x-Position der Queen

    for (unsigned ty = 0; ty < qy; ++ty) {
        if (board[ty] == qx) { // Prüfe ober- und unterhalb
            return false;
        } else if (ty-board[ty] == qy-qx) { //prüfe diagonal links oben
            return false;
        } else if (ty+board[ty] == qy+qx) { // prüfe diagonal rechts oben
            return false;
        }
    }
    return true;
}

/**
 * @brief Gibt das Schachbrett board auf der Console aus.
 * 
 * @param board Repräsentation des Schachbretts.
 */
void printBoard(int board[8]) {
    cout << ++count << endl;
    for (unsigned y = 0; y < 8; ++y) {
        for (int x = 0; x < 8; ++x) {
            cout << (x == board[y] ? "X " : "o ");
        }
        cout << endl;
    }
    cout << endl << endl;
}

/**
 * @brief Rekursive Implementation des Damenproblems.
 * 
 * @param board Repräsentation des Schachbretts als eindimensionales int-Array. Werte repräsentieren die x-Koordinate der Queen, -1 repräsentiert keine gesetzte Queen
 * @param y Aktuelle zu prüfende y-Koordinate, sollte beginnend bei 0 starten
 * @return int Wurde eine korrekte Lösung gefunden, so wird 0 returnt, ansonsten -1
 */
void queens(int board[8], int y) {
    if (y == 8) {
        printBoard(board);
    } else {
        for (unsigned x = 0; x < 8; ++x) {
            if (check(board, x, y)) {
                board[y] = x;

                queens(board, y+1) ;
            }
        }
    }
}

int main() {
    int board[8];
    queens(board, 0);
    return 0;
}