/**
 * @file aufgabe02zusatz.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Schreiben Sie eine Methode, die alle Lösung des 8-Dame Problems ausgibt. Das 8-Dame Problem besteht darin, auf einem Schachbrett (8x8 Feld) 8 Damen derart zu platzieren, dass keine Dame eine andere Dame schlagen kann, d.h. in keiner Zeile und in keiner Spalte und in keiner Diagonalen dürfen 2 oder mehr Damen gleichzeitig stehen. (Tipp: arbeiten Sie rekursiv) → Speichern Sie das Schachbrett in nur einem unsigned Integer, anstelle eines unsigned-Arrays!
 * @version 0.1
 * @date 2022-05-14
 * 
 * @copyright Copyright (c) 2022
 */

#include <iostream>
// #include <bitset>
using namespace std;

unsigned count = 0; 

/**
 * @brief Gibt den aktuellen x-Wert der Queen in Zeile y zurück 
 * 
 * @param board unsigned-Repräsentation des Schachbretts
 * @param y y-Position der Queen
 * @return unsigned x-Position der Queen
 */
unsigned getQueen (unsigned board, unsigned y) {
    // unsigned queen = board >> y*3 & 7;
    // cout << endl << bitset<3*8>(board).to_string() << endl;
    // cout << endl << bitset<3*8>(queen).to_string() << endl;
    return (board >> (y*3)) & 7;
}

/**
 * @brief Setzt die Queen auf dem board in Zeile y, Spalte x
 * 
 * @param board unsigned-Repräsentation des Schachbretts
 * @param y y-Position der Queen
 * @param x x-Position der Queen
 */
void setQueen (unsigned& board, unsigned y, unsigned x) {
    /* unsigned mask = 0; // Maske generieren
    for (unsigned i = 8; i--;) {
        mask = mask << 3;
        if (i != y) {
            mask += 7;
        }
    } 
    // board = board & mask; // Alte x-Position an Zeile y zurücksetzen */
    board = board & ~(7 << (y * 3));

    //x = x << y * 3; 
    board = board + (x << y * 3); // Neue x-Position an Zeile y setzen
}

/**
 * @brief Überprüft, ob eine Queen an der Stelle qx, qy auf dem Schachbrett board legal gestellt werden kann. Dabei wird nur diagonal oberhalb und vertikal ober- und unterhalb geprüft (heißt, die Positionen der Queens sollten von oben nach unten gesetzt/geprüft werden)
 * 
 * @param board Repräsentation des Schachbretts als eindimensionales int-Array
 * @param qx Zu prüfende x-Koordinate der Queen
 * @param qy Zu prüfende y-Koordinate der Queen
 * @return true Die geprüfte Position ist legal
 * @return false Die geprüfte Position ist nicht legal
 */
bool check(unsigned board, unsigned qx, unsigned qy) {
    // ty = zu testendes y
    // board[ty] = zu testendes x
    // qy = y-Position der Queen
    // qx = x-Position der Queen

    for (unsigned ty = 0; ty < qy; ++ty) {
        if (getQueen(board, ty) == qx) { // Prüfe ober- und unterhalb
            return false;
        } else if (ty-getQueen(board, ty) == qy-qx) { //prüfe diagonal links oben
            return false;
        } else if (ty+getQueen(board, ty) == qy+qx) { // prüfe diagonal rechts oben
            return false;
        }
    }
    return true;
}

/**
 * @brief Gibt das Schachbrett board auf der Console aus.
 * 
 * @param board Repräsentation des Schachbretts.
 */
void printBoard(unsigned board) {
    cout << ++count << "    " << board << endl;
    for (unsigned y = 0; y < 8; ++y) {
        for (unsigned x = 0; x < 8; ++x) {
            cout << (x == getQueen(board, y) ? "X " : "o ");
        }
        cout << endl;
    }
    cout << endl << endl;
}

/**
 * @brief Rekursive Implementation des Damenproblems.
 * 
 * @param board Repräsentation des Schachbretts als eindimensionales int-Array. Werte repräsentieren die x-Koordinate der Queen, -1 repräsentiert keine gesetzte Queen
 * @param y Aktuelle zu prüfende y-Koordinate, sollte beginnend bei 0 starten
 * @return int Wurde eine korrekte Lösung gefunden, so wird 0 returnt, ansonsten -1
 */
void queens(unsigned board, int y) {
    if (y == 8) {
        printBoard(board);
    } else {
        for (unsigned x = 0; x < 8; ++x) {
            if (check(board, x, y)) {
                setQueen(board, y, x);

                queens(board, y+1) ;
            }
        }
    }
}

int main() {
    unsigned board = 0;
    queens(board, 0);
    // getQueen(342391,0);
    // unsigned board = 0;
    // setQueen(board, 5, 7);
    return 0;
}