/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie den Shell- und Quicksort für int-Arrays. Testen Sie Ihre Programme mit Arrays der Größen 10, 100, 1000 und 10000. Dabei sollen die Arrays bereits sortiert, invers sortiert und mit zufälligen Zahlen belegt sein.
 * @version 0.1
 * @date 2022-05-24
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <cstdlib>
#include <chrono>

using namespace std;
typedef chrono::steady_clock sclk;

/**
 * @brief Füllt das übergebene int-Array mit zufälligen Werten.
 * 
 * @param field 
 * @param size 
 */
void fillRandom (int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        field[i] = rand();
    }
}

/**
 * @brief Füllt das übergebene int-Array mit sortierten Werten. 
 * 
 * @param field 
 * @param size 
 */
void fillSorted (int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        field[i] = i;
    }
}

/**
 * @brief Füllt das übergebene int-Array mit rückwerts sortierten Werten.
 * 
 * @param field 
 * @param size 
 */
void fillSortedBack (int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        field[i] = size-i;
    }
}

/**
 * @brief Tauscht die Werte an den Stellen i1 und i2 im übergebenen int-Array aus
 * 
 * @param field 
 * @param i1 
 * @param i2 
 */
void swap(int* field, unsigned i1, unsigned i2) {
    int temp = field[i1];
    field[i1] = field[i2];
    field[i2] = temp;
}

void quickSortHelp (int* field, int iLeft, int iRight) {
    const int MID = field[(iLeft + iRight) / 2];
    // const int MID = field[(iLeft)];
    int l = iLeft;
    int r = iRight;

    while (l < r) {
        while (field[l] < MID) { ++l; }
        while (MID < field[r]) { --r;}

        if (l <= r) {
            swap(field, l++, r--);
        }
    }
    if (iLeft < r) { quickSortHelp(field, iLeft, r); }
    if (iRight > l) { quickSortHelp(field, l, iRight); }
}

/**
 * @brief Sortiert das übergebene int-Array mit dem Quick-Sort-Algorithmus.
 * 
 * @param field 
 * @param size 
 */
void quickSort (int* field, unsigned size) {
    quickSortHelp(field, 0, size-1);
}

/**
 * @brief Sortiert das übergebene int-Array mit dem Shell-Sort-Algorithmus.
 * 
 * @param field 
 * @param size 
 */
void heapSort (int*& field, unsigned size) {
    unsigned iDist = 1;
    for ( ; iDist <= size / 9; iDist = 3 * iDist + 1);
    for ( ; iDist > 0; iDist /= 3) {
        for (unsigned i1 = iDist; i1 < size; ++i1) {
            const int IVAL = field[i1];
            unsigned i2 = i1;
            while (i2 >= iDist && IVAL < field[i2-iDist]) {
                field[i2] = field[i2-iDist];
                i2 = i2 - iDist;
            }
            field[i2] = IVAL;
        }
    }

}

/**
 * @brief Gibt das übergebene int-Array auf der Console aus.
 * 
 * @param field 
 * @param size 
 */
void printArray (const int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        cout << field[i] << endl;
    }
    cout << endl;
}

typedef void (*fillFunction) (int* field, unsigned size);

void singleRunQuick (unsigned i, fillFunction f) {
    int* field = new int[i];
        (*f)(field, i);
        // printArray(field, i);
        sclk::time_point startTime = sclk::now();
        quickSort(field, i);
        sclk::time_point endTime = sclk::now();
        // printArray(field, i);

        sclk::duration durationTime = endTime - startTime;
        cout << "Dauer für Quicksort Größe " << i << " in µSec.: " << chrono::duration_cast<chrono::microseconds>(durationTime).count() << endl;   
        delete[] field;
}

void singleRunShell (unsigned i, fillFunction f) {
    int* field = new int[i];
        (*f)(field, i);
        // printArray(field, i);
        sclk::time_point startTime = sclk::now();
        heapSort(field, i);
        sclk::time_point endTime = sclk::now();
        // printArray(field, i);

        sclk::duration durationTime = endTime - startTime;
        cout << "Dauer für Shellsort Größe " << i << " in µSec.: " << chrono::duration_cast<chrono::microseconds>(durationTime).count() << endl;   
        delete[] field;
}

int main () {
    for (unsigned i = 10; i <= 10000; i*=10) {
        singleRunQuick(i, &fillRandom);
        singleRunShell(i, &fillRandom);
        singleRunQuick(i, &fillSorted);
        singleRunShell(i, &fillSorted);
        singleRunQuick(i, &fillSortedBack);
        singleRunShell(i, &fillSortedBack);
    }

    return 0;
}