/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie den Heap- und Mergesort für int-Arrays. Testen Sie Ihre Programme mit Arrays der Größen 10, 100, 1000 und 10000. Dabei sollen die Arrays bereits sortiert, invers sortiert und mit zufälligen Zahlen belegt sein.
 * @version 0.1
 * @date 2022-06-13
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <cstdlib>
#include <chrono>

using namespace std;
typedef chrono::steady_clock sclk;

struct Heap {
    unsigned m_iNext;
    int* m_Keys;

    Heap (unsigned uiSize) {
        m_iNext = 0;
        m_Keys = new int[uiSize];
    }

    Heap (unsigned uiSize, int* field) {
        m_iNext = 0;
        m_Keys = field;
    }

    ~Heap () {
        delete[] m_Keys;
        // cout << "tot" << endl;
    }

    void downheap (unsigned uEnd, unsigned uIndex) {
        int k = m_Keys[uIndex];
        while (uIndex < uEnd / 2) {
            unsigned uSon = 2 * uIndex + 1;
            if (uSon < uEnd - 1 && m_Keys[uSon] < m_Keys[uSon + 1])
                ++uSon;
            if (!(k < m_Keys[uSon]))
                break;
            m_Keys[uIndex] = m_Keys[uSon];
            uIndex = uSon;
        }
        m_Keys[uIndex] = k;
    }
};

/**
 * @brief Füllt das übergebene int-Array mit zufälligen Werten.
 * 
 * @param field 
 * @param size 
 */
void fillRandom (int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        field[i] = rand();
    }
}

/**
 * @brief Füllt das übergebene int-Array mit sortierten Werten. 
 * 
 * @param field 
 * @param size 
 */
void fillSorted (int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        field[i] = i;
    }
}

/**
 * @brief Füllt das übergebene int-Array mit rückwerts sortierten Werten.
 * 
 * @param field 
 * @param size 
 */
void fillSortedBack (int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        field[i] = size-i;
    }
}

/**
 * @brief Tauscht die Werte an den Stellen i1 und i2 im übergebenen int-Array aus
 * 
 * @param field 
 * @param i1 
 * @param i2 
 */
void swap(int* field, unsigned i1, unsigned i2) {
    int temp = field[i1];
    field[i1] = field[i2];
    field[i2] = temp;
}

void mergeSortHelp (int* field, unsigned iLeft, unsigned iRight, int* tmp) {
    if (iLeft < iRight) {
        const unsigned MIDDLE = (iLeft + iRight) / 2;
        mergeSortHelp(field, iLeft, MIDDLE, tmp);
        mergeSortHelp(field, MIDDLE + 1, iRight, tmp);

        unsigned tmp_length = iRight - iLeft + 1;
        // int* tmp = new int[tmp_length];
        for (unsigned i = iLeft; i <= MIDDLE; ++i) {
            tmp[i-iLeft] = field[i];
        }
        for (int i = MIDDLE + 1; i <= iRight; ++i) {
            tmp[tmp_length-i+MIDDLE] = field[i];
        }
        unsigned iL = 0;
        unsigned iR = tmp_length - 1;
        for (unsigned i = iLeft; i <= iRight; ++i) {
            field[i] = tmp[iL] < tmp[iR] ? tmp[iL++] : tmp [iR--];
        }
        // delete[] tmp;
    }
}

/**
 * @brief Sortiert das übergebene int-Array mit dem Merge-Sort-Algorithmus.
 * 
 * @param field 
 * @param size 
 */
void mergeSort (int* field, unsigned size) {
    int* fieldHelp = new int[size];
    mergeSortHelp(field, 0, size-1, fieldHelp);
    delete[] fieldHelp;
}

/**
 * @brief Sortiert das übergebene int-Array mit dem Heap-Sort-Algorithmus.
 * 
 * @param field 
 * @param size 
 */
void heapSort (Heap& heap, unsigned size) {
    // Heap heap(size, field);
    for (unsigned i = ((size - 1) - 1) / 2; i--; ) {
        heap.downheap(size, i+1);
    }
    for (unsigned i = size - 1; i--; ) {
        swap(heap.m_Keys, 0, i+1);
        heap.downheap(i+1, 0);
    }
}

/**
 * @brief Gibt das übergebene int-Array auf der Console aus.
 * 
 * @param field 
 * @param size 
 */
void printArray (const int* field, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        cout << field[i] << endl;
    }
    cout << endl;
}

typedef void (*fillFunction) (int* field, unsigned size);

void singleRunMerge (unsigned i, fillFunction f) {
    int* field = new int[i];
        (*f)(field, i);
        // printArray(field, i);
        sclk::time_point startTime = sclk::now();
        mergeSort(field, i);
        sclk::time_point endTime = sclk::now();
        // printArray(field, i);

        sclk::duration durationTime = endTime - startTime;
        cout << "Dauer für Mergesort Größe " << i << " in mSec.: " << chrono::duration_cast<chrono::milliseconds>(durationTime).count() << endl;   
        delete[] field;
}

void singleRunHeap (unsigned i, fillFunction f) {
    int* field = new int[i];
        (*f)(field, i);
        // printArray(field, i);
        sclk::time_point startTime = sclk::now();

        Heap heap(i, field);
        heapSort(heap, i);

        sclk::time_point endTime = sclk::now();
        // printArray(field, i);

        sclk::duration durationTime = endTime - startTime;
        cout << "Dauer für Heapsort Größe " << i << " in mSec.: " << chrono::duration_cast<chrono::milliseconds>(durationTime).count() << endl;   
}

int main () {        
    // singleRunHeap(10, &fillSortedBack);

    for (unsigned i = 10; i <= 10000000; i*=10) {
        // singleRunMerge(i, &fillRandom);
        singleRunHeap(i, &fillRandom);
        // singleRunMerge(i, &fillSorted);
        singleRunHeap(i, &fillSorted);
        // singleRunMerge(i, &fillSortedBack);
        singleRunHeap(i, &fillSortedBack);
    } 

    return 0;
}