/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie eine Vektorklasse für double Zahlen. Fügen Sie der Implementierung eine push_front Methode hinzu, die am Anfang des Vektors ein neues Element hinzufügt
 * @version 0.1
 * @date 2022-06-15
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>
using namespace std;

class MyVector {
    public:
        /**
         * @brief Construct a new My Vector object
         * 
         * @param initialCapacity The initial capacity of the vector (default is 1)
         * @param capacityIncrement The value the capacity of the vector is added, if the capacity limit is reached (default is doubleing the capacity) 
         */
        MyVector (unsigned initialCapacity = 1, unsigned capacityIncrement = 0) :
            m_capacityIncrement(capacityIncrement),
            m_capacity(initialCapacity),
            m_size(0)
        {
            m_arr = new double[initialCapacity];
        }

        ~MyVector () {
            delete[] m_arr;
            // cout << "Im dead" << endl;
        }

        /**
         * @brief saves the value at the end of the vector
         * 
         * @param value the value to save
         */
        void push_back (double value) {
            if (m_size >= m_capacity) {
                resize();
            }
            m_arr[m_size++] = value;
        }

        /**
         * @brief saves the value at the front of the vector
         * 
         * @param value the value to save
         */
        void push_front (double value) {
            if (m_size >= m_capacity) {
                resize();
            }
            for (unsigned i = m_size; i > 0; --i) {
                m_arr[i] = m_arr[i-1];
            }
            m_arr[0] = value;
            ++m_size;
        }

        /**
         * @brief returns the value of the index in the vector. If index is out of bounds, it returns NULL.
         * 
         * @param index index in the vector
         * @return double value of the index in the vector
         */
        double get (unsigned index) const {
            assert(index < m_size);
            return m_arr[index];
            
        }

        /**
         * @brief prints the content of the vector to the console.
         * 
         */
        void print () const {
            for (unsigned i = 0; i < m_capacity; ++i) {
                cout << "[" << i << "] = " << m_arr[i] << endl;
            }
        }

        /**
         * @brief Returns the size of the vector
         * 
         * @return unsigned 
         */
        unsigned size () const {
            return m_size-1;
        }

    private: 
        /**
         * @brief Increment the capacity of the vector according to the rules defined during creation of the vector.
         * 
         */
        void resize () {
            // cout << "resize()" << endl;
            unsigned newCapacity = m_capacity + (m_capacityIncrement == 0 ? m_capacity : m_capacityIncrement);
            double* temp = new double[newCapacity];
            for (unsigned i = 0; i < m_capacity; ++i) {
                temp[i] = m_arr[i];
            }
            delete[] m_arr;
            m_arr = temp;
            m_capacity = newCapacity;
        }

        double* m_arr;
        const unsigned m_capacityIncrement;
        unsigned m_capacity;
        unsigned m_size;
};

int main () {
    MyVector mv;
    for (unsigned i = 0; i < 10; ++i) {
        mv.push_back(i+0.1);
        //mv.print();
        //cout << endl << endl;
    }

    for (unsigned i = 0; i < 10; ++i) {
        mv.push_front(i+0.9);
    }

    mv.print();
    return 0;
}