/**
 * @file aufgabe02.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Modifizieren Sie Ihre push_front Methode derart, dass sie analog wie die push_back Methode auch am Anfang mehr Elemente alloziiert als notwendig ist.
 * @version 0.1
 * @date 2022-06-15
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>
using namespace std;

class MyVector {
    public:
        /**
         * @brief Construct a new My Vector object
         * 
         * @param initialCapacity The initial capacity of the vector (default is 1)
         * @param capacityIncrement The value the capacity of the vector is added, if the capacity limit is reached (default is doubleing the capacity) 
         */
        MyVector (unsigned initialCapacity = 1, unsigned capacityIncrement = 0) :
            m_phyStart(0),
            m_size(0),
            m_capacityIncrement(capacityIncrement),
            m_capacity(initialCapacity)
        {
            m_arr = new double[initialCapacity];
        }

        ~MyVector () {
            delete[] m_arr;
            // cout << "Im dead" << endl;
        }

        /**
         * @brief saves the value at the end of the vector
         * 
         * @param value the value to save
         */
        void push_back (double value) {
            if (m_size >= (m_capacity-m_phyStart)) {
                resize(true);
            }
            m_arr[m_phyStart+m_size] = value;
            ++m_size;
        }

        /**
         * @brief saves the value at the front of the vector
         * 
         * @param value the value to save
         */
        void push_front (double value) {
            if (m_phyStart == 0) {
                resize(false);
            }
            m_arr[--m_phyStart] = value;
            ++m_size;
        }

        /**
         * @brief returns the value of the index in the vector. If index is out of bounds, it returns NULL.
         * 
         * @param index index in the vector
         * @return double value of the index in the vector
         */
        double get (unsigned index) {
            assert(index < m_size);
            return m_arr[index+m_phyStart];
            
        }

        /**
         * @brief prints the content of the vector to the console.
         * 
         */
        void print () {
            for (unsigned i = 0; i < m_capacity; ++i) {
                cout << "[" << i << "] = " << m_arr[i] << endl;
            }
        }

        /**
         * @brief Returns the size of the vector
         * 
         * @return unsigned 
         */
        unsigned size () {
            return m_size;
        }
    private: 
        /**
         * @brief Increment the capacity of the vector according to the rules defined during creation of the vector.
         * @param back Add capacity to the end of the vector (if false: add capacity to the front of the vector)
         * 
         */
        void resize (bool back) {
            unsigned newCapacity = m_capacity + (m_capacityIncrement == 0 ? m_capacity : m_capacityIncrement);
            double* temp = new double[newCapacity];
            unsigned offset = back ? m_phyStart : (m_capacityIncrement == 0 ? m_capacity : m_capacityIncrement);
            for (unsigned i = 0; i < size(); ++i) {
                    temp[i+offset] = get(i);
            }
            if (!back) {
                m_phyStart = offset;
            }
            delete[] m_arr;
            m_arr = temp;
            m_capacity = newCapacity;
        }

        double* m_arr;
        const unsigned m_capacityIncrement;
        unsigned m_capacity;
        unsigned m_size;
        unsigned m_phyStart;
};

int main () {
    MyVector mv;
    for (unsigned i = 0; i < 10; ++i) {
        mv.push_back(i+0.1);
        mv.print();
        cout << endl << endl;
    }
    for (unsigned i = 0; i < 10; ++i) {
        mv.push_front(i+0.9);
        mv.print();
        cout << endl << endl;
    }

    for (unsigned i = 0; i < mv.size(); ++i) {
        cout << i << ": " << mv.get(i) << endl;
    }
    return 0;
}