/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie eine Klasse A, die einen int Wert speichert. Implementieren Sie eine einfach verkettete Liste, die A Objekte speichert. Legen Sie eine Liste an, speichern mehrere A Objekte ab und geben die Liste aus. Die Liste soll zusätzlich die Methoden front (erstes Element zurückliefern) und back (letztes Element zurückliefern) besitzen. Denken Sie an den Destruktor, Copykonstruktor, Zuweisungsoperator und Ausgabeoperator.
 * @version 0.1
 * @date 2022-06-25
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>

using namespace std;

/**
 * @brief Wrapper-class for an integer
 * 
 */
class A {
    public:
    A (int value) :
        m_value(value)
    { }

    friend ostream& operator<< (ostream& os, const A& crArg) {
        os << crArg.m_value;
        return os;
    }

    private:
    int m_value;
};

/**
 * @brief Element of a list
 * 
 */
class Element {
    public: 
    /**
     * @brief Construct a new Element object
     * 
     * @param value Value to save
     * @param next Next element in the list
     */
    Element (A value, Element* next) :
        m_value(value),
        m_next(next)
    { }
    
    /**
     * @brief Destroy the Element object
     * 
     */
    ~Element () { }

    /**
     * @brief returns the value of the object
     * @return value of the object
     * 
     */
    A getValue () const {
        return m_value;
    }

    /**
     * @brief Sets the value to the provided new value
     * 
     */
    void setValue (A value) {
        m_value = value;
    }

    /**
     * @brief Set the Next object in the list
     * 
     * @param next Pointer to the new next object in the List
     */
    void setNext (Element* next) {
        m_next = next;
    }

    /**
     * @brief returns the next element in the list
     * @return the next element in the list; nullptr, if this ist the last element in the list
     * 
     */
    Element* getNext () const {
        return m_next;
    }

    private: 
    A m_value;
    Element* m_next;
};

class SingleList {
    public: 
    SingleList () = default;

    /**
     * @brief Destroy the Single List object
     * 
     */
    ~SingleList () {
        Element* p = m_head;
        while (p != nullptr) {
            Element* next_p = p->getNext();
            delete p;
            p = next_p;
        }
        cout << "dead" << endl;
    }

    /**
     * @brief Copy-construct a new Single List object
     * 
     * @param crArg Object to be copied
     */
    SingleList (const SingleList &crArg) :
        m_head(nullptr)
    {
        Element* help = nullptr;
        for (Element* p = crArg.m_head; p != nullptr; p = p->getNext()) {
            Element* tmp = new Element (p->getValue(), nullptr);
            if (help == nullptr) {
                m_head = tmp;
            } else {
                help->setNext(tmp);
            }
            help = tmp;
        }
    }

    /**
     * @brief Adds a new object to the front of the list
     * @param value Value, that should be added to the list
     * 
     */
    void pushFront (int value) {
        m_head = new Element(value, m_head);
    }
   
    /**
     * @brief returns first element of the list.
     * 
     */
    Element& front () const {
        return *m_head;
    }

    /**
     * @brief returns last element of the list.
     * 
     */
    Element& back () const {
        Element* p = m_head;
        while (p->getNext() != nullptr) {
            p = p->getNext();
        }
        return *p;
    }

    /**
     * @brief Prints the list to the console
     * 
     */
    friend ostream& operator<< (ostream& os, const SingleList& crArg) {
        unsigned index = 0;
        Element* p = crArg.m_head;
        while (p != nullptr) {
            cout << "[" << index << "] = " << (p->getValue()) << endl;
            p = p->getNext();
            ++index;
        }
        return os;
    }

    /**
     * @brief Zuweisungsoperator
     * 
     * @param crArg 
     * @return SingleList& 
     */
    SingleList& operator= (SingleList crArg) {
        Element *tmp = crArg.m_head;
        crArg.m_head = m_head;
        m_head = tmp;
        return *this;
    }

    private: 
    Element* m_head = nullptr;
};

int main () {
    {
        SingleList sl;
        for (int i = -3; i < 4; ++i) {
            sl.pushFront(i);
        }
        cout << sl << endl;
        SingleList sl2 = sl;
        sl.front().setValue(-5);
        sl.back().setValue(5);
        cout << sl << endl;
        cout << sl2 << endl;
    }

    return 0;
}