/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Erweitern Sie die Implementierungen von Vektoren und Listen um eine pop_front und pop_back Methode, die das erste bzw. letzte Element entfernt. 
 * @version 0.1
 * @date 2022-06-25
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>

using namespace std;

class Element {
    public: Element (int value, Element* next) :
        m_value(value),
        m_next(next)
    {

    }
    
    public: ~Element () {

    }

    /**
     * @brief returns the value of the object
     * @return value of the object
     * 
     */
    public: int getValue () const {
        return m_value;
    }

    /**
     * @brief Sets the value to the provided new value
     * 
     */
    public: void setValue (int value) {
        m_value = value;
    }

    /**
     * @brief returns the next element in the list
     * @return the next element in the list; nullptr, if this ist the last element in the list
     * 
     */
    public: Element* getNext () const {
        return m_next;
    }

    public: void setNext (Element* next) {
        m_next = next;
    }

    private: int m_value;
    private: Element* m_next;
};

class SingleList {
    public: SingleList (Element* front = nullptr) : 
        m_head(front)  {

    }

    public: ~SingleList () {
        Element* p = m_head;
        Element* next_p = m_head->getNext();
        while (p != nullptr) {
            next_p = p->getNext();
            delete p;
            p = next_p;
        }
        cout << "dead" << endl;
    }

    public: SingleList (const SingleList &crArg) :
        m_head(new Element(crArg.front()->getValue(), nullptr))
    {
        Element* p = crArg.front()->getNext();
        while (p != nullptr) {
            pushFront(p->getValue());
            p = p->getNext();
        }
    }

    /**
     * @brief Adds a new object to the front of the list
     * @param value Value, that should be added to the list
     * 
     */
    public: void pushFront (int value) {
        
        m_head = new Element(value, m_head);
    }

    public: void pop_front () {
        Element* old_head = m_head;
        m_head = m_head->getNext();
        delete old_head;
    }

    public: void pop_back() {
        Element* p = m_head;
        while (p->getNext()->getNext() != nullptr) {
            p = p->getNext();
        }
        delete p->getNext();
        p->setNext(nullptr);
    }

    /**
     * @brief returns first element of the list.
     * 
     */
    public: Element* front () const {
        return m_head;
    }

    /**
     * @brief returns last element of the list.
     * 
     */
    public: Element* back () const {
        Element* p = m_head;
        while (p->getNext() != nullptr) {
            p = p->getNext();
        }
        return p;
    }

    /**
     * @brief Prints the list to the console
     * 
     */
    public: void printList () const {
        unsigned index = 0;
        Element* p = m_head;
        while (p != nullptr) {
            cout << "[" << index << "] = " << (p->getValue()) << endl;
            p = p->getNext();
            ++index;
        }
        cout << endl;
    }

    private: Element* m_head;
};

class MyVector {
    public:
        /**
         * @brief Construct a new My Vector object
         * 
         * @param initialCapacity The initial capacity of the vector (default is 1)
         * @param capacityIncrement The value the capacity of the vector is added, if the capacity limit is reached (default is doubleing the capacity) 
         */
        MyVector (unsigned initialCapacity = 1, unsigned capacityIncrement = 0) :
            m_phyStart(0),
            m_size(0),
            m_capacityIncrement(capacityIncrement),
            m_capacity(initialCapacity)
        {
            m_arr = new double[initialCapacity];
        }

        ~MyVector () {
            delete[] m_arr;
            // cout << "Im dead" << endl;
        }

        /**
         * @brief saves the value at the end of the vector
         * 
         * @param value the value to save
         */
        void push_back (double value) {
            if (m_size >= (m_capacity-m_phyStart)) {
                resize(true);
            }
            m_arr[m_phyStart+m_size] = value;
            ++m_size;
        }

        /**
         * @brief saves the value at the front of the vector
         * 
         * @param value the value to save
         */
        void push_front (double value) {
            if (m_phyStart == 0) {
                resize(false);
            }
            m_arr[--m_phyStart] = value;
            ++m_size;
        }

        void pop_front () {
            ++m_phyStart;
            --m_size;
        }

        void pop_back () {
            --m_size;
        }

        /**
         * @brief returns the value of the index in the vector. If index is out of bounds, it returns NULL.
         * 
         * @param index index in the vector
         * @return double value of the index in the vector
         */
        double get (unsigned index) {
            assert(index < m_size);
            return m_arr[index+m_phyStart];
            
        }

        /**
         * @brief prints the content of the vector to the console.
         * 
         */
        void print () {
            for (unsigned i = 0; i < m_size; ++i) {
                cout << "[" << i << "] = " << m_arr[m_phyStart+i] << endl;
            }
            cout << endl;
        }

        /**
         * @brief Returns the size of the vector
         * 
         * @return unsigned 
         */
        unsigned size () {
            return m_size;
        }
    private: 
        /**
         * @brief Increment the capacity of the vector according to the rules defined during creation of the vector.
         * @param back Add capacity to the end of the vector (if false: add capacity to the front of the vector)
         * 
         */
        void resize (bool back) {
            unsigned newCapacity = m_capacity + (m_capacityIncrement == 0 ? m_capacity : m_capacityIncrement);
            double* temp = new double[newCapacity];
            unsigned offset = back ? m_phyStart : (m_capacityIncrement == 0 ? m_capacity : m_capacityIncrement);
            for (unsigned i = 0; i < size(); ++i) {
                    temp[i+offset] = get(i);
            }
            if (!back) {
                m_phyStart = offset;
            }
            delete[] m_arr;
            m_arr = temp;
            m_capacity = newCapacity;
        }

        double* m_arr;
        const unsigned m_capacityIncrement;
        unsigned m_capacity;
        unsigned m_size;
        unsigned m_phyStart;
};

int main () {
    // SingleList-Test
    SingleList sl = SingleList();
    MyVector v = MyVector();
    for (int i = -3; i < 4; ++i) {
        sl.pushFront(i);
        v.push_front(i);
    }
    sl.printList();
    sl.pop_back();
    sl.pop_front();
    sl.printList();
    
    v.print();
    v.pop_back();
    v.pop_front();
    v.print();

    return 0;
}