/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie eine Klasse Rational zur Speicherung und Darstellung rationaler Zahlen. Implementieren Sie sinnvolle Konstruktoren und Operatoren (+, -, *, /, <, <=, >, >=, ==, !=, <<, =). Die Speicherung soll immer in gekürzter Form erfolgen. 
 * @version 0.1
 * @date 2022-07-03
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>

using namespace std;

/**
 * @brief Speichert eine rationale Zahl
 * 
 */
class Rational {
    public:
    /**
     * @brief Construct a new Rational object
     * 
     * @param zaehler Zähler des Bruchs
     * @param nenner Nenner des Bruchs
     */
    Rational (int zaehler, unsigned nenner = 1) :
    mi_zaehler(zaehler),
    mu_nenner(nenner)
    {
        kuerzen();
    } 

    friend Rational operator+ (const Rational& crArg1, const Rational& crArg2);
    friend Rational operator- (const Rational& crArg1, const Rational& crArg2);
    friend Rational operator* (const Rational& crArg1, const Rational& crArg2);
    friend Rational operator/ (const Rational& crArg1, const Rational& crArg2);
    friend bool operator<  (const Rational& crArg1, const Rational& crArg2);
    friend bool operator<= (const Rational& crArg1, const Rational& crArg2);
    friend bool operator>  (const Rational& crArg1, const Rational& crArg2);
    friend bool operator>= (const Rational& crArg1, const Rational& crArg2);
    friend bool operator== (const Rational& crArg1, const Rational& crArg2);
    friend bool operator!= (const Rational& crArg1, const Rational& crArg2);
    friend ostream& operator<< (ostream&, const Rational& crArg);
    void operator= (const Rational& crArg) {
        mi_zaehler = crArg.mi_zaehler; 
        mu_nenner = crArg.mu_nenner;
    }

    private:
    /**
     * @brief Berechnet den größten gemeinsamen Teiler (ggT) der Zahlen a und b und gibt diesen wieder zurück
     * 
     * @param a Erste Zahl
     * @param b Zweite Zahl
     * @return int größter gemeinsamer Teiler der Zahlen a und b.
     */
    int ggt (int a, int b) {
        int h = (a > b) ? b : a;
        for (int i = h; i > 1; --i) {
            if ((a % i) == 0 && (b % i) == 0) {
                return i;
            }
        }
        return 1;
    }

    /**
     * @brief Kürzt den Bruch
     * 
     */
    void kuerzen () {
        int ggt0 = ggt(mi_zaehler, mu_nenner);
        mi_zaehler /= ggt0;
        mu_nenner /= ggt0;
    }

    int mi_zaehler;
    unsigned mu_nenner;
};

Rational operator+ (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mu_nenner + crArg2.mi_zaehler * crArg1.mu_nenner, crArg1.mu_nenner * crArg2.mu_nenner);
}

Rational operator- (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mu_nenner - crArg2.mi_zaehler * crArg1.mu_nenner, crArg1.mu_nenner * crArg2.mu_nenner);
}

Rational operator* (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mi_zaehler, crArg1.mu_nenner * crArg2.mu_nenner);
}

Rational operator/ (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mu_nenner, crArg1.mu_nenner * crArg2.mi_zaehler);
}

bool operator< (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner < crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator<= (const Rational& crArg1, const Rational& crArg2) {
    //return (crArg1.mi_zaehler * crArg2.mu_nenner <= crArg2.mi_zaehler * crArg1.mu_nenner);
    return !(crArg2 < crArg1);
}

bool operator> (const Rational& crArg1, const Rational& crArg2) {
    //return (crArg1.mi_zaehler * crArg2.mu_nenner > crArg2.mi_zaehler * crArg1.mu_nenner);
    return crArg2 < crArg1;
}

bool operator>= (const Rational& crArg1, const Rational& crArg2) {
    //return (crArg1.mi_zaehler * crArg2.mu_nenner >= crArg2.mi_zaehler * crArg1.mu_nenner);
    return !(crArg1 < crArg2);
}

bool operator== (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner == crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator!= (const Rational& crArg1, const Rational& crArg2) {
    //return (crArg1.mi_zaehler * crArg2.mu_nenner != crArg2.mi_zaehler * crArg1.mu_nenner);
    return !(crArg1 == crArg2);
}

ostream& operator<< (ostream& os, const Rational& crArg) {
    os << "(" << crArg.mi_zaehler << " / " << crArg.mu_nenner << ")";
    return os;
}

int main () {
    Rational r1 = Rational(5, 25);
    Rational r2 = Rational(-2, 7);
    cout << r1 << endl;
    cout << r2 << endl;
    cout << r1 + r2 << endl;
    cout << r1 - r2 << endl;
    cout << r1 * r2 << endl;
    cout << r1 / r2 << endl;
    cout << (r1 < r2 ? "True" : "False") << endl;
    cout << (r1 <= r2 ? "True" : "False") << endl;
    cout << (r1 == r2 ? "True" : "False") << endl;
    cout << (r1 != r2 ? "True" : "False") << endl;
    cout << (r1 > r2 ? "True" : "False") << endl;
    cout << (r1 >= r2 ? "True" : "False") << endl;
    r1 = r2;
    cout << r1 << "  " << r2 << endl;
    cout << r1 + 5 << endl;
    cout << 5 + r1 << endl;
    return 0;
}