/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie den Bubblesort für Rationale Zahlen. 
 * @version 0.1
 * @date 2022-07-03
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>

using namespace std;

/**
 * @brief Speichert eine rationale Zahl
 * 
 */
class Rational {
    public:
    /**
     * @brief Construct a new Rational object
     * 
     * @param zaehler Zähler des Bruchs
     * @param nenner Nenner des Bruchs
     */
    Rational (int zaehler, unsigned nenner = 1) :
    mi_zaehler(zaehler),
    mu_nenner(nenner)
    {
        kuerzen();
    } 

    friend Rational operator+ (const Rational& crArg1, const Rational& crArg2);
    friend Rational operator- (const Rational& crArg1, const Rational& crArg2);
    friend Rational operator* (const Rational& crArg1, const Rational& crArg2);
    friend Rational operator/ (const Rational& crArg1, const Rational& crArg2);
    friend bool operator<  (const Rational& crArg1, const Rational& crArg2);
    friend bool operator<= (const Rational& crArg1, const Rational& crArg2);
    friend bool operator>  (const Rational& crArg1, const Rational& crArg2);
    friend bool operator>= (const Rational& crArg1, const Rational& crArg2);
    friend bool operator== (const Rational& crArg1, const Rational& crArg2);
    friend bool operator!= (const Rational& crArg1, const Rational& crArg2);
    friend ostream& operator<< (ostream&, const Rational& crArg);
    void operator= (const Rational& crArg) {
        mi_zaehler = crArg.mi_zaehler; 
        mu_nenner = crArg.mu_nenner;
    }

    private:
    /**
     * @brief Berechnet den größten gemeinsamen Teiler (ggT) der Zahlen a und b und gibt diesen wieder zurück
     * 
     * @param a Erste Zahl
     * @param b Zweite Zahl
     * @return int größter gemeinsamer Teiler der Zahlen a und b.
     */
    int ggt (int a, int b) {
        int h = (a > b) ? b : a;
        for (int i = h; i > 1; --i) {
            if ((a % i) == 0 && (b % i) == 0) {
                return i;
            }
        }
        return 1;
    }

    /**
     * @brief Kürzt den Bruch
     * 
     */
    void kuerzen () {
        int ggt0 = ggt(mi_zaehler, mu_nenner);
        mi_zaehler /= ggt0;
        mu_nenner /= ggt0;
    }

    int mi_zaehler;
    unsigned mu_nenner;
};

Rational operator+ (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mu_nenner + crArg2.mi_zaehler * crArg1.mu_nenner, crArg1.mu_nenner * crArg2.mu_nenner);
}

Rational operator- (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mu_nenner - crArg2.mi_zaehler * crArg1.mu_nenner, crArg1.mu_nenner * crArg2.mu_nenner);
}

Rational operator* (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mi_zaehler, crArg1.mu_nenner * crArg2.mu_nenner);
}

Rational operator/ (const Rational& crArg1, const Rational& crArg2) {
    return Rational(crArg1.mi_zaehler * crArg2.mu_nenner, crArg1.mu_nenner * crArg2.mi_zaehler);
}

bool operator< (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner < crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator<= (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner <= crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator> (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner > crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator>= (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner >= crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator== (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner == crArg2.mi_zaehler * crArg1.mu_nenner);
}

bool operator!= (const Rational& crArg1, const Rational& crArg2) {
    return (crArg1.mi_zaehler * crArg2.mu_nenner != crArg2.mi_zaehler * crArg1.mu_nenner);
}

ostream& operator<< (ostream& os, const Rational& crArg) {
    os << "(" << crArg.mi_zaehler << " / " << crArg.mu_nenner << ")";
    return os;
}

/**
 * @brief Tauscht zwei Elemente im Array an den Indizes i1 und i2.
 * 
 * @param field Array, indem die zwei Elemente getauscht werden sollen.
 * @param i1 Index des ersten Elements.
 * @param i2 Index des zweiten Elements.
 */
void swap(Rational* field, unsigned i1, unsigned i2) {
    Rational temp = field[i1];
    field[i1] = field[i2];
    field[i2] = temp;
}

/**
 * @brief Sortiert das übergebene Array mittels des Bubble-Sort-Algorithmus.
 * 
 * @param arr Zu sortierendes Array.
 * @param size Größe des zu sortierenden Arrays.
 */
void bubbleSort(Rational* arr, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        bool b = false;
        for (unsigned j = 0; j < size-(i+1); ++j) {
            if (arr[j] > arr[j+1]) {
                swap(arr, j+1, j);
                b = true;
            }
        }
        if (!b) {
            return;
        }
    }
    return;
}

void printArr (Rational* arr, unsigned size) {
    for (unsigned i = 0; i < size; ++i) {
        cout << arr[i] << "  ";
    }
    cout << endl;
}

int main () {
    Rational arr[5] = {Rational(1, 5), Rational(5), Rational(22, 11), Rational(2, 7), Rational(42, 35)};
    printArr(arr, 5);
    bubbleSort(arr, 5);
    printArr(arr, 5);

    return 0;
}