/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie eine Klasse, die gleichzeitig einen Stack und eine Queue für unsigned int realisiert. Diese Klasse soll von der Klasse der doppelt verketteten Liste abgeleitet sein. Welche der beiden Funktionalitäten im tatsächlichen Objekt realisiert werden, soll durch ein boolesches Flag im Konstruktor entschieden werden. 
 * @version 0.1
 * @date 2022-07-03
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>

using namespace std;

class A {
    public:
    A (int value) :
        m_value(value)
    { }

    friend ostream& operator<< (ostream& os, const A& crArg) {
        os << crArg.m_value;
        return os;
    }

    private:
    int m_value;
};

class Element {
    public: Element (A value, Element* next, Element* prev) :
        m_value(value),
        m_next(next),
        m_prev(prev)
    {
        if (m_next != nullptr) {
            m_next->setPrev(this);
        }
        if (m_prev != nullptr) {
            m_prev->setNext(this);
        }
    }
    
    public: ~Element () {

    }

    /**
     * @brief returns the value of the object
     * @return value of the object
     * 
     */
    public: A getValue () const {
        return m_value;
    }

    /**
     * @brief Sets the value to the provided new value
     * 
     */
    public: void setValue (A value) {
        m_value = value;
    }

    /**
     * @brief returns the next element in the list
     * @return the next element in the list; nullptr, if this ist the last element in the list
     * 
     */
    public: Element* getNext () const {
        return m_next;
    }

    public: void setNext (Element* next) {
        m_next = next;
    }

    public: Element* getPrev () const {
        return m_prev;
    }

    public: void setPrev (Element* prev) {
        m_prev = prev;
    }

    private: A m_value;
    private: Element* m_next;
    private: Element* m_prev;
};

class DoubleList {
    public: DoubleList () = default;

    public: ~DoubleList () {
        Element* p = m_head;
        while (p != nullptr) {
            Element* next_p = p->getNext();
            delete p;
            p = next_p;
        }
        cout << "dead" << endl;
    }

    public: DoubleList (const DoubleList &crArg) {
        // TODO: Copy-Konstruktor anpassen
        Element* p = crArg.m_head;
        while (p != nullptr) {
            pushBack(p->getValue());
            p = p->getNext();
        }
    }

    /**
     * @brief Adds a new object to the front of the list
     * @param value Value, that should be added to the list
     * 
     */
    public: void pushFront (A value) {
        m_head = new Element(value, m_head, nullptr);
        if (m_tail == nullptr) {
            m_tail = m_head;
        }
    }

    public: void pop_front () {
        Element* old_head = m_head;
        m_head = m_head->getNext();
        delete old_head;
        m_head->setPrev(nullptr);
    }
   
    /**
     * @brief returns first element of the list.
     * 
     */
    public: Element* front () const {
        return m_head;
    }

    public: void pushBack (A value) {
        m_tail = new Element(value, nullptr, m_tail);
        if (m_head == nullptr) {
            m_head = m_tail;
        }
    }

    public: void pop_back() {
        Element* old_tail = m_tail;
        m_tail = m_tail->getPrev();
        delete old_tail;
        m_tail->setNext(nullptr);
    }

    /**
     * @brief returns last element of the list.
     * 
     */
    public: Element& back () const {
        return *m_tail;
    }

    /**
     * @brief Prints the list to the console
     * 
     */
    public: void printList () const {
        unsigned index = 0;
        Element* p = m_head;
        while (p != nullptr) {
            cout << "[" << index << "] = " << (p->getValue()) << endl;
            p = p->getNext();
            ++index;
        }
        cout << endl;
    }

    public: void printListBack () const {
        unsigned index = 0;
        Element* p = m_tail;
        while (p != nullptr) {
            cout << "[" << index << "] = " << (p->getValue()) << endl;
            p = p->getPrev();
            ++index;
        }
        cout << endl;
    }

    private: Element* m_head = nullptr;
    private: Element* m_tail = nullptr;
};

class StackOrQueue {
    public:
    StackOrQueue (bool isStack) :
    mb_isStack(isStack)
    {

    }

    private:
    const bool mb_isStack;
};