/**
 * @file aufgabe02.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie für den binären Suchbaum über unsigned int eine rekursive Suchmethode und eine Methode, die die Anzahl der Knoten zurückliefert.
 * @version 0.1
 * @date 2022-07-08
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>
using namespace std;

class Node {
    public:
        /**
         * @brief Construct a new Node object
         * 
         * @param key Key-Value
         * @param data Data-Value
         */
        Node (unsigned key, unsigned data) :
        mu_key(key),
        mu_data(data)
        {

        }

        /**
         * @brief Copys a Node object
         * 
         * @param crArg 
         */
        Node (const Node& crArg) :
        mu_key(crArg.mu_key),
        mu_data(crArg.mu_data),
        mo_left(crArg.mo_left != nullptr ? new Node (*crArg.mo_left) : nullptr),
        mo_right(crArg.mo_right != nullptr ? new Node (*crArg.mo_right) : nullptr)
        {
        }

        /**
         * @brief Destroy the Node object
         * 
         */
        ~Node () {
            // Delete auf Nullptr gibt keinen Fehler
            delete mo_left;
            delete mo_right;
        }

        /**
         * @brief Counts the size of the node and its subnotes
         * 
         * @return unsigned Number of nodes
         */
        unsigned size () const {
            countChilds(this);
        }
        

        /**
         * @brief Get the Key value
         * 
         * @return unsigned Key-value
         */
        unsigned getKey () const {
            return mu_key;
        }

        /**
         * @brief Get the next Left-node object
         * 
         * @return Node
         */
        Node* getLeft () const {
            return mo_left;
        }

        /**
         * @brief Set the next Left-node object
         * 
         * @param left 
         */
        void setLeft (Node* left) {
            mo_left = left;
        }

        /**
         * @brief Get the next Right-node object
         * 
         * @return Node
         */
        Node* getRight () const {
            return mo_right;
        }

        /**
         * @brief Set the next Right-node object
         * 
         * @param right 
         */
        void setRight (Node* right) {
            mo_right = right;
        }

        Node& operator= (const Node crArg) = delete;

        friend ostream& operator<< (ostream&, const Node& crArg);

    private:
        /**
         * @brief Count the number of nodes starting at n
         * 
         * @param n Start node
         * @return unsigned Number of nodes
         */
        static unsigned countChilds (const Node* n) {
            if (n == nullptr) {
                return 0;
            } else {
                return 1 + countChilds(n->getLeft()) + countChilds(n->getRight());
            }
        }

        /**
         * @brief Prints the space infront of a node-entry.
         * TODO: correct the lines
         * 
         * @param os Output stream
         * @param versatz Number of whitespaces
         */
        void printVersatz (ostream& os, unsigned versatz) const {
            for (unsigned i = 0; i < versatz; ++i) {
                os << (i % 2 == 0 ? "│" : " ");
            }
        }

        /**
         * @brief Prints the node and its childs to os
         * 
         * @param os Output stream
         * @param versatz Number of whitespaces
         * @param ebene Not in use
         */
        void printNode (ostream& os, unsigned versatz, unsigned ebene) const {
            os << mu_data << endl;
            printVersatz(os, versatz);
            os << "├─";
            if (mo_left != nullptr) {
                mo_left->printNode(os, versatz+2, ebene+1);
            }
            os << endl;
            printVersatz(os, versatz);
            os << "└─";
            if (mo_right != nullptr) {
                mo_right->printNode(os, versatz+2, ebene+1);
            }
        }
        
        const unsigned mu_key;
        const unsigned mu_data;
        Node* mo_left = nullptr;
        Node* mo_right = nullptr;
};

ostream& operator<< (ostream& os, const Node& crArg) {
    crArg.printNode(os, 0, 0);
    return os;
}

class BinTree {
    public:
        BinTree () = default;

        /**
         * @brief Copys a Node object
         * 
         * @param crArg 
         */
        BinTree (const BinTree& crArg) :
        mo_root(crArg.mo_root != nullptr ? new Node (*crArg.mo_root) : nullptr) 
        {

        }

        /**
         * @brief Destroy the Bin Tree object
         * 
         */
        ~BinTree () {
            delete mo_root;
        }

        /**
         * @brief Searches in the binary tree for the node with the key parameter.
         * If a node is found, the pointer to the node is returned.
         * If no node is found, a nullpointer will be returned.
         * 
         * @param key Key to be searched
         * @return Node* Pointer to the node object. Is null, if object can not be found
         */
        Node* search (unsigned key) const {
            recSearch (key, mo_root);
        }

        /**
         * @brief Inserts a new node object in the binary tree
         * 
         * @param key Key-value of the new node
         * @param data Data-value of the new node
         */
        void insert (unsigned key, unsigned data) {
            mo_root = insert (mo_root, key, data);
        }

        /**
         * @brief Counts the size of the binary tree
         * 
         * @return unsigned Number of nodes
         */
        unsigned size () const {
            if (mo_root != nullptr) {
                return mo_root->size();
            }
            return 0;
        }

        /**
         * @brief Get the Root node object of the binary tree
         * 
         * @return Node* Root node object
         */
        Node* getRoot () const {
            return mo_root;
        }

        BinTree& operator= (BinTree crArg) {
            Node* tmp = crArg.mo_root;
            crArg.mo_root = mo_root;
            mo_root = tmp;
            return *this;
        }

        friend ostream& operator<< (ostream&, const Node& crArg);

    private:
        /**
         * @brief Searches in the binary tree for the node with the key parameter.
         * If a node is found, the pointer to the node is returned.
         * If no node is found, a nullpointer will be returned.
         * 
         * @param key Key to be searched
         * @return Node* Pointer to the node object. Is null, if object can not be found
         */
        Node* recSearch (unsigned key, Node* n) const {
            if (n == nullptr) {
                return nullptr;
            } else {
                if (key < n->getKey()) {
                    return recSearch(key, n->getLeft());
                } else if (n->getKey() < key) {
                    return recSearch(key, n->getRight());
                } else {
                    return n;
                }
            }
        }
        
        /**
         * @brief Inserts a new node object in the binary tree
         * 
         * @param n Start node
         * @param key Key-value of the new node
         * @param data Data-value of the new node
         * @return Node* Returns the new inserted node
         */
        Node* insert (Node* n, unsigned key, unsigned data) {
            if (n == nullptr) { // kann durch if (n) ersetzt werden, weil nullptr == false?
                return new Node(key, data);
            } else {
                if (key < n->getKey()) {
                    n->setLeft(insert(n->getLeft(), key, data));
                } else if (key > n->getKey()) {
                    n->setRight(insert(n->getRight(), key, data));
                }
                return n;
            }
        }

        Node* mo_root = nullptr;

};

ostream& operator<< (ostream& os, const BinTree& crArg) {
    os << crArg.getRoot();
    return os;
}

int main () {
    BinTree bt = BinTree();
    bt.insert(34, 34);
    bt.insert(13, 13);
    bt.insert(40, 40);
    bt.insert(5, 5);
    bt.insert(17, 17);
    bt.insert(38, 38);
    bt.insert(42, 42);
    bt.insert(15, 15);
    bt.insert(16, 16);
    cout << bt << endl << endl;
    
    cout << bt.size() << endl; // 9
    cout << bt.getRoot()->getRight()->size() << endl; // 3
    cout << bt.getRoot()->getRight()->getRight()->size() << endl << endl; // 1

    cout << bt.search(15) << endl;
    cout << bt.search(55) << endl;

    return 0;
}